from collections import defaultdict

from .sandhi import SandhiRule


class StemmingRuleSet:

    def __init__(self):
        # mapping of key to list of rules
        self.key_to_rules = defaultdict(list)
        # a reverse mapping of surface to key, stem pairs
        self.surface_to_key_stem = defaultdict(set)

    def add(self, key, rule, tags=None):
        r = SandhiRule(rule, tags)
        self.key_to_rules[key].append(r)
        self.surface_to_key_stem[r.surface].add((key, r))
        return r

    def inflect(self, stem, key, tag_filter=None):
        base_endings = []
        default = []
        tag_filter = tag_filter or set()

        for rule in self.key_to_rules[key]:
            skip = False
            for tag in rule.tags:
                if tag[0] == "+" and tag[1:] not in tag_filter:
                    skip = True
                    break
                if tag[0] == "-" and tag[1:] in tag_filter:
                    skip = True
                    break
            if skip:
                continue

            base = rule.match_theme(stem)
            if base is not None:
                if rule.stem:
                    base_endings.append((base, rule.distinguisher, rule))
                else:
                    default.append((base, rule.distinguisher, rule))

        # only use default if there are no other options
        if len(base_endings) == 0 and len(default) > 0:
            used_default = True
            base_endings = default
        else:
            used_default = False

        for base, ending, rule in base_endings:
            yield {
                "base": base,
                "ending": ending,
                "rule": rule,
                "used_default": used_default,
            }

    def possible_stems(self, form):
        for surface, rules in self.surface_to_key_stem.items():
            if form.endswith(surface):
                for key, rule in rules:
                    yield (key, form[:-len(surface)] + rule.stem)



import re


class Inflexion:

    def __init__(self):
        self.lexicons = []
        self.stemming_rule_sets = []

    def add_lexicon(self, lexicon):
        self.lexicons.append(lexicon)

    def add_stemming_rule_set(self, stemming_rule_set):
        self.stemming_rule_sets.append(stemming_rule_set)

    def generate(self, lemma, key, tag_filter=None):

        tag_filter = tag_filter or set()

        stems = set()
        for lexicon in self.lexicons:
            stems.update(lexicon.find_stems(lemma, key, tag_filter))

        results = set()
        for stem in stems:
            for stemming_rule_set in self.stemming_rule_sets:
                for result in stemming_rule_set.inflect(stem, key, tag_filter):
                    results.add(result["base"] + result["ending"])

        return results

    def parse(self, form):
        results = set()

        for stemming_rule_set in self.stemming_rule_sets:
            for key, stem in stemming_rule_set.possible_stems(form):
                for lexicon in self.lexicons:
                    for lemma, key_regex, tags in \
                            lexicon.stem_to_lemma_key_regex[stem]:
                        if re.match(key_regex, key):
                            results.add((lemma, key))

        return results



from setuptools import setup

setup(
    name="inflexion",
    version="0.4",
    description="generic code for morphological analysis",
    license="MIT",
    url="http://github.com/jtauber/inflexion",
    author="James Tauber",
    author_email="jtauber@jtauber.com",
    packages=["inflexion"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Topic :: Text Processing",
        "Topic :: Text Processing :: Linguistic",
    ],
)



import re


class SandhiRule:

    def __init__(self, rule, tags=None):
        """
        rule is string of form A|B>C<D|E
        tags is set of strings annotating the rule for later filtering, etc.
        """
        self.tags = tags or set()
        self.a, bcd, self.e = rule.split("|")
        self.b, cd = bcd.split(">")
        self.c, self.d = cd.split("<")
        self.theme = self.a
        self.stem = self.a + self.b
        self.suffix = self.d + self.e
        self.distinguisher = self.c + self.e
        self.surface = self.a + self.c + self.e

    def __repr__(self):
        if self.tags:
            return "SandhiRule('{0.a}|{0.b}>{0.c}<{0.d}|{0.e}', " \
                "tags={1})".format(self, self.tags)
        else:
            return "SandhiRule('{0.a}|{0.b}>{0.c}<{0.d}|{0.e}')".format(self)

    def match_theme(self, stem):
        """
        If the given stem matches this rule's stem part, return the theme
        (which may be more than this rule's theme part if this rule's stem part
        is only the rightmost part of the given stem) or return None if stems
        don't match.
        """
        if re.match(".*" + self.stem + "$", stem):
            if self.b:
                return stem[:-len(self.b)]
            else:
                return stem
        else:
            return None



from collections import defaultdict
import re


class Lexicon:

    def __init__(self):
        # mapping of lemma to list of (key_regex, stem, tags)
        self.lemma_to_stems = defaultdict(list)
        # a reverse mapping of stem to set of (lemma, key_regex, tags)
        self.stem_to_lemma_key_regex = defaultdict(set)

    def add(self, lemma, key_regex, stem, tags=None):
        tags = tags or set()
        self.lemma_to_stems[lemma].append((key_regex, stem, tags))
        self.stem_to_lemma_key_regex[stem].add(
            # we use tuple(sorted(...)) to make deterministically hashable
            (lemma, key_regex, tuple(sorted(tags))))

    def find_stems(self, lemma, key, tag_filter=None):
        """
        returns a (possibly empty) stem_set for the given lemma and key
        """
        tag_filter = tag_filter or set()

        prev_key_regex = None
        result = set()

        for key_regex, stem, tags in self.lemma_to_stems[lemma]:
            skip = False
            for tag in tags:
                if tag[0] == "+" and tag[1:] not in tag_filter:
                    skip = True
                    break
                if tag[0] == "-" and tag[1:] in tag_filter:
                    skip = True
                    break
            if skip:
                continue

            if re.match(key_regex, key):
                if key_regex != prev_key_regex:  # this means multiple stems
                    result = set()               # for same key_regex must be
                    prev_key_regex = key_regex   # contiguously added

                result.add(stem)

        return result
